
% June 19, 2009

% Force the Appending of semi-colons

Isis_Append_Semicolon=0;

% Make sure filenames and background file names get printed in list
% functions 
Isis_List_Filenames=1;

% Set the minimum allowable statistical error to a tiny number

Minimum_Stat_Err = 1.e-30;

putenv("PGPLOT_BACKGROUND=white"); % invert default for better view.
putenv("PGPLOT_FOREGROUND=black");

% A message function I'll use a lot below

public define msg(str_array)
{
   () = printf("\n");
   foreach(str_array)
   {
      variable str = ();
      () = printf(" %s\n", str);
   }
   () = printf("\n");
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Some useful aliasing (or simple functions akin to aliasing)

alias("fit_fun","model");  % Bowing to my XSPEC days ...
alias("fit_counts","fit"); % Bowing to my XSPEC days ...

public define newpar()     % Still Further bowing to my XSPEC days ...
{
   variable idx, value, freeze, min, max;
   switch(_NARGS)
   {
    case 2:
    (idx,value) = ();
    set_par(idx,value);
    return;
   }
   {
    case 3:
    (idx,value,freeze) = ();
    if(freeze < 0){ freeze = 1; } % This is the XSPEC part
    set_par(idx,value,freeze);
    return;
   }
   {
    case 4:
    (idx,value,freeze,min) = ();
    if(freeze < 0){ freeze = 1; } % This is the XSPEC part
    set_par(idx,value,freeze,min);
    return;
   }
   {
    case 5:
    (idx,value,freeze,min,max) = ();
    if(freeze < 0){ freeze = 1; } % This is the XSPEC part
    set_par(idx,value,freeze,min,max);
    return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% Fitting and Confidence Functions, last Updated: March 30, 2008

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Public Functions in This File.  Usage message (almost always) when 
% function is called without arguments.

% sep_grid           : Set the eval method to SEPARATE_GRID
% mrg_grid           : Set the eval method to MERGED_GRID
% usr_grid           : Set up logarithmically spaced USER_GRIDs
% fstat_mod          : Set the fit statistic to sigma=model
% fstat_gehr         : Set the fit statistic to sigma=gehrels
% fstat_norm         : Set the fit statistic to sigma=data
% lmdif              : Set the fit method to lmdif 
% minim              : Set the fit method to minim
% subplex            : Set the fit method to subplex
% marq               : Set the fit method to marquardt
% conf_map_fail_hook : Switch between minim/lmdif when doing
%                      error contours, and the fit fails
% conf_map           : Call conf_map_counts with conf_map_fail_hook
% add_syst           : The new rebinning error hook function, if one 
%                      sets systematic errors or background corrections
% set_systematic     : Set systematic errors on the data
% corback            : Set the data background to have a fittable 
%                      normalization, using the back_fun utility
% fcorback_fit       : The fit function used by corback
% corfile            : In addition to the data background, set another
%                      background subtracted file as a background with
%                      fittable normalization
% fcorfile_fit       : The fit function used by corfile

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Public Variables That Affect This File.  

% Isis_Reverse_Channels : Determines whether systematic error channel
%                         numbers refer to energy or wavelength
% syst_indx_lo          : Stores lower bin boundaries for sys. errors
% syst_indx_hi          : Stores upper bin boundaries for sys. errors
% syst_indx_err         : Storres systematic errors (as fractions). 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sep_grid()
{
   switch(_NARGS)
   {
    case 1:
      variable a = ();
      set_eval_grid_method(SEPARATE_GRID,a);
      return;
   }
   {
      variable str = [" sep_grid(a);"," ",
      "   Equivalent to set_eval_grid_method(SEPARATE_GRID,a);",""];
      msg(str);
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%

public define fstat_mod()
{
      set_fit_statistic("chisqr;sigma=model");
      return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%

public define fstat_gehr()
{
      set_fit_statistic("chisqr;sigma=gehrels");
      return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%

public define fstat_norm()
{
      set_fit_statistic("chisqr;sigma=data");
      return;
}

alias("fstat_norm","fstat_data");
alias("fstat_norm","fstat_dat");

%%%%%%%%%%%%%%%%%%%%%%%%

public define mrg_grid()
{
   switch(_NARGS)
   {
    case 1:
      variable a = ();
      set_eval_grid_method(MERGED_GRID,a);
      return;
   }
   {
      variable str = [" sep_grid(a);"," ",
      "   Equivalent to set_eval_grid_method(MERGED_GRID,a);",""];
      msg(str);
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static variable lgh_lb = Float_Type[30], lgh_ub = @lgh_lb, lgh_scl = @lgh_lb;

public define log_grid_hook(id,s)
{
    s.bin_lo = _A(10^[lgh_lb[id]:lgh_ub[id]:lgh_scl[id]]);
    s.bin_hi = make_hi_grid(s.bin_lo);
    return s;
}


public define usr_grid()
{ 
   variable a,lb,ub,scl,cache,i,id;
   switch(_NARGS)
   {
    case 4:
      (a,lb,ub,scl) = ();
      cache = 0;
   }
   {
    case 5:
      (a,lb,ub,scl,cache) = ();
      if(cache!=0){cache = 1;}
   }
   {
      variable str = [" usr_grid(a,lb,ub,scl [,cache]);"," ",
      "   Create a USER_GRID for each dataset, a[i], such that there is a",
      "   logarithmic energy grid running over keV energies:","",
      "      10^[lb[i]:ub[i]:scl[i]]","",
      "   Called via set_eval_grid_method(USER_GRID, a, &log_grid_hook [,cache]);",
      "   *Also* sets the kernel to: set_kernel(a, \"std;eval=all\");","",
      " Inputs:",
      "   a    : Vector indices for which the USER_GRID will be applied",
      "   lb   : Vector of log10 of lower keV bounds",
      "   ub   : Vector of log10 of upper keV bounds",
      "   scl  : Uniform log step size of the grid",
      "   cache: (Optional) If non-zero, cache the model evaluation",""];
      msg(str);
      return;
   }

   a = [a]; lb = [lb]; ub = [ub]; scl = [scl];

   variable la = length(a), lbz = lb[0], ubz = ub[0], sclz = scl[0];
   variable llb = length(lb);
   variable lub = length(ub);
   variable lscl = length(scl);

   if(cache==0)
   {
      if(llb < la)
      {
         lb = [lb,ones(la-llb)*lbz];
      }
      if(lub < la)
      {
         ub = [ub,ones(la-lub)*ubz];
      }
      if(lscl < la)
      {
         scl = [scl,ones(la-lscl)*sclz];
      }
   }
   else
   {
      lb = Float_Type[la]; ub = @lb; scl = @lb;
      lb[*] = lbz;
      ub[*] = ubz;
      scl[*] = sclz;
   }

   i = 0;
   foreach(a)
   {
      id = ();
      lgh_lb[id] = lb[i];
      lgh_ub[id] = ub[i];
      lgh_scl[id] = scl[i];
      i++;
   }

   set_eval_grid_method (USER_GRID, a, &log_grid_hook, cache);
   set_kernel(a, "std;eval=all");
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%

public define subplex()
{
   set_fit_method("subplex");
   return;
}

%%%%%%%%%%%%%%%%%%%%

public define marq()
{
   set_fit_method("marquardt");
   return;
}

%%%%%%%%%%%%%%%%%%%%

public define lmdif()
{
   set_fit_method("lmdif");
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define conf_map_fail_hook(p1, p2, best_pars, try_pars,fit_info)
{
   variable save_method = get_fit_method ();
                  
   variable meth = strchop(save_method,';',0)[0];

   if(meth=="marquardt" or meth=="lmdif")
   {
      subplex;
   }
   else
   {
      lmdif;
   }

   () = fit_counts (&fit_info);
                  
   set_fit_method (save_method);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static variable conf_map_info=struct{fail};
conf_map_info.fail = &conf_map_fail_hook;

%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define conf_map(x,y)
{
   return conf_map_counts(x,y,conf_map_info);
}

% My own public variable toggle to say that in certain situations
% (e.g., systematic error application), reverse the indices (i.e.,
% go from wavelength space to energy space)

public variable Isis_Reverse_Channels = 1;

% Right now only ready for systematic errors on 12 simultaneous
% data sets.  Edit if you need more!

public variable syst_indx_lo = Array_Type[13];
public variable syst_indx_hi = Array_Type[13];
public variable syst_indx_err = Array_Type[13];
public variable syst_indx_bkg = Array_Type[13];
public variable syst_indx_bsc = Float_Type[13];
public variable data_indx_bkg = Array_Type[13];
public variable data_indx_err = Array_Type[13];
public variable data_indx_bsc = Float_Type[13];

public variable syst_indx_app = Float_Type[13];
syst_indx_app[*]=0;

public variable back_indx_app = Float_Type[13];
back_indx_app[*]=0;

public variable corback_comp=Integer_Type[13];
corback_comp[*]=0;

public variable data_indx_app=Integer_Type[13];
data_indx_app[*]=0;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define add_syst(orig_cts, orig_stat_err, grouping)
{
   variable systerr, lorig, indx_lngth, i, indces, syst_in, syst_ones;
   variable indx_lo, indx_hi, syst_errs;

   % Get # of counts per grouped bin 
   systerr = rebin_array(orig_cts,grouping);

   % Default systematic errors are 0%, since this is also used in
   % background correction statistics

   lorig = length(orig_cts);
   syst_in = Float_Type[lorig];
   syst_in[*]=0;
   syst_ones = ones(lorig);

   if(Isis_Verbose > 0){ print(""); }

   if(syst_indx_app[Isis_Active_Dataset]==1)
   {
      indx_lngth = length(syst_indx_lo[Isis_Active_Dataset]);
 
      % Isis does things in wavelength channels, but for things like
      % RXTE, I prefer keV, and want to add systematics via keV channels

      if(Isis_Reverse_Channels)
      {
         indx_lo = lorig - 1 - reverse(syst_indx_hi[Isis_Active_Dataset]);
         indx_hi = lorig - 1 - reverse(syst_indx_lo[Isis_Active_Dataset]);
         syst_errs=reverse(syst_indx_err[Isis_Active_Dataset]);
            
         if(Isis_Verbose > 0)
         { 
            print(" Applying systematic errors to energy channels.");
         }
      }
      else
      {
         indx_lo = syst_indx_lo[Isis_Active_Dataset];
         indx_hi = syst_indx_hi[Isis_Active_Dataset];
         syst_errs = syst_indx_err[Isis_Active_Dataset];

         if(Isis_Verbose > 0)
         { 
            print(" Applying systematic errors to wavelength channels.");
         }
      }

      i = 0;
      loop( indx_lngth )
      {
         % No error checking to make sure these are sensible!

         indces = [indx_lo[i]:indx_hi[i]];
         syst_in[indces] = syst_errs[i];
         i++;
      }
   }

   % Get the average systematic errors in grouped bins    
   syst_in = rebin_array(syst_in,grouping) /
             rebin_array(syst_ones,grouping);

   variable meth = strtok(get_fit_statistic,";=");
   meth = meth[length(meth)-1];

   % The new error is added in quadrature to the old error
   if( meth=="gehrels" )
   {
      systerr = (1. + sqrt(systerr+0.75))^2 + (syst_in*systerr)^2; 
    
      if(Isis_Verbose > 1)
      {
         print(" Errors applied with Gehrels statistics");
      }
   }
   else
   {
      systerr = (1. + syst_in^2 * systerr) * systerr;

      if(Isis_Verbose > 1)
      {
         print(" Errors applied with Poisson statistics");
      }
   }
  
   % Add in the saved data background
   systerr += rebin_array(syst_indx_bkg[Isis_Active_Dataset],grouping)
              *syst_indx_bsc[Isis_Active_Dataset];

   % Add in the extra correction file background, if it exists
   if(data_indx_app[Isis_Active_Dataset] == 1)
   {
      systerr += (data_indx_err[Isis_Active_Dataset])^2;

      if(Isis_Verbose > 1)
      {
         print(" Additional background errors applied.");
      }
   }

   if(Isis_Verbose > 1){ print(""); }

   return sqrt(systerr);
}       

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define create_back(hist_index)
{
   rebin_data(hist_index,0);

   variable b = get_back(hist_index);

   if( b==NULL )
   {
      syst_indx_bkg[hist_index] = 
      Float_Type[length(get_data_counts(hist_index).bin_lo)];
   }
   else
   {
      syst_indx_bkg[hist_index] = b;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define create_backscale(hist_index)
{    
   variable scl;
   variable dscl = get_data_backscale(hist_index)[0];
   variable dexp = get_data_exposure(hist_index);

   variable bfile = get_data_info(hist_index).bgd_file;
   if(bfile != NULL and bfile != "#_define_bgd()" and bfile != "")
   {
      variable bexp = fits_read_key(bfile,"EXPOSURE");
      variable bscl = fits_read_key(bfile,"BACKSCAL");

      if(bexp == NULL or bscl == NULL or bexp == 0 or bscl == 0)
      {
         scl = 1.;
      }
      else 
      {
         scl = (dexp*dscl)/(bexp*bscl);
      }
   }
   else
   {
      scl = 1;
   }
   return scl;
}   

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define set_systematic()
{
   variable hist_index,binlo,binhi,err;
   variable set_sys, set_not;

   switch(_NARGS)
   {
    case 1:
      (hist_index) = ();
      syst_indx_app[abs(hist_index)] = 0;

      if( corback_comp[abs(hist_index)] == 0 )
      {
         set_rebin_error_hook(abs(hist_index),NULL);
         back_indx_app[abs(hist_index)] = 0;
         data_indx_app[abs(hist_index)] = 0;
      }

      set_sys=get_data_info(abs(hist_index)).rebin;
      set_not=get_data_info(abs(hist_index)).notice_list;
   }
   {
    case 4:
      (hist_index,binlo,binhi,err) = ();

      binlo = [binlo];
      binhi = [binhi];
      err = [err];

      set_sys=get_data_info(abs(hist_index)).rebin;
      set_not=get_data_info(abs(hist_index)).notice_list;

      if(hist_index < 0)
      {
         syst_indx_app[abs(hist_index)] = 0;

         if( corback_comp[abs(hist_index)] == 0 )
         {
            set_rebin_error_hook(abs(hist_index),NULL);
            back_indx_app[abs(hist_index)] = 0;
            data_indx_app[abs(hist_index)] = 0;
         }
      }
      else
      {
         if( ( length(binlo) != length(binhi) ) or 
             ( length(binlo) != length(err) ) or
             ( length(binlo) < 1 ) )
         {
            message(" Inconsistent array lengths -- ");
            message("    No systematic errors implemented");
            return;
         }

         syst_indx_lo[hist_index] = binlo;
         syst_indx_hi[hist_index] = binhi;
         syst_indx_err[hist_index] = err;

         if( back_indx_app[hist_index] == 0 )
         {
           create_back(hist_index);
           syst_indx_bsc[hist_index] = create_backscale(hist_index);
           back_indx_app[hist_index] = 1;
         }

         set_rebin_error_hook(hist_index,"add_syst");
         syst_indx_app[hist_index] = 1;
      }
   }
   {
      variable str = [" set_systematic[s](hist_index,[binlo],[binhi],[err]);"," ",
       "   Turns on systematic errors on data set `hist_index'",
       "   over channel numbers bounded by [binlo]:[binhi] (inclusive),",
       "   with a fractional level of [err]"," ",
       "   Note that *averaged* systematic errors are applied *after*",
       "   rebinning, and are then added in quadrature to Poisson or Gehrels.",
       "   ",
       "   Setting hist_index to < 0 removes the systematic",
       "   errors for data set abs(hist_index)"," ",
       "   Giving only hist_index, but no other input, also removes",
       "   the systematic errors for data set hist_index","",
       "   WARNING: The usual Poisson, or Gehrels modified, errors will",
       "    be applied, depending upon what was in place *at the time*",
       "    the systematic errors were applied, *or* at the time of any",
       "    subsequent rebinning.  If you're unsure, rebin!"];

      variable str_chan;

      if(Isis_Reverse_Channels)
      {
         str_chan = [" ",
         " Note:  You have Isis_Reverse Channels variable set such that",
         " input channel numbers here are referring to *energy* channels"];
      }
      else
      {
         str_chan = [" ",
         " Note:  You have Isis_Reverse Channels variable set such that",
         " input channel numbers here are referring to *wavelength* channels"];
      }

      msg([str,str_chan]);
      return;
   }

   % Make sure the systematics are applied (else it won't happen
   % until the data gets rebinned)  

   rebin_data(abs(hist_index),set_sys);
   ignore(abs(hist_index));
   notice_list(abs(hist_index),set_not);

   return;
}

alias("set_systematic","set_systematics");
alias("set_systematic","setsystematics");
alias("set_systematic","setsystematic");


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Defining fit function for the background %%%

public define fcorback_fit(lo,hi,par)
{
   return par[0]*syst_indx_bkg[int(par[1])];
}
	
add_slang_function("fcorback",["norm","id"]);

% The id parameter of fcorback should always be frozen to an 
% integer value.  It's default value is 2 but it should be 
% set to be the id of the dataset it describes and remain 
% frozen there. The default maximum can be changed if necessary to accomodate
% more than 13 simultaneous datasets.

%%%%%%%%%%%%%%%%%%%%%%%%%%

define fcorback_default(i)
{
   switch(i)
   {
    case 0:
    return(1,0,0,2);
   }
   {
    case 1:
    return(1,1,1,13);
   }
}

set_param_default_hook("fcorback","fcorback_default");

%%%%%%%%%%%%%%%%%%%%%%%

public define corback()
{
   variable hist_index,b,rb,n;
   switch(_NARGS)
   {
    case 1:
      hist_index=();

      variable ahist = abs(int(hist_index));
      rb=get_data_info(ahist).rebin;
      n=get_data_info(ahist).notice_list;
      
      if(hist_index < 0)
      {
         back_fun(ahist,NULL);
         if( syst_indx_app[ahist] == 0 )
         {
            set_rebin_error_hook(ahist,NULL);
            back_indx_app[ahist] = 0;
         }
         corback_comp[ahist] = 0;
         data_indx_app[ahist] = 0;
         rebin_data(ahist,rb);
         ignore(ahist);
         notice_list(ahist,n);
         return;
      }

      % Wiping out the binning of the dataset and getting the unbinned 
      % background.  Then wiping out the background.
	      
      back_fun(ahist,NULL);

      if( back_indx_app[ahist] == 0 )
      {
        create_back(ahist);
        syst_indx_bsc[ahist] = create_backscale(ahist);
        back_indx_app[ahist] = 1;
      }

      corback_comp[ahist]=1;

      set_rebin_error_hook(ahist,"add_syst");
   
      rebin_data(ahist,rb);
      ignore(ahist);
      notice_list(ahist,n);

      back_fun(ahist,"fcorback("+string(ahist)+")");    
      set_par("fcorback("+string(ahist)+").id",ahist);

      return;
   }
   {
      variable str = [" corback(hist_index);"," ",
       "    Sets the background associated with data set `hist_index' as a correction",
       "    file which can be scaled via a constant using the automatically defined",
       "    function `fcorback(hist_index)'.  The background errors continue to be",
       "    added in quadrature (either with Poisson or Gehrels statistics - ",
       "    whichever was in place the last time the data was rebinned), but",
       "    without accounting for the background fit constant.","",
       "    To clear the background correction do:","", 
       "       isis> corback(-hist_index)"];
      msg(str);
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define fcorfile_fit(lo,hi,par)
{
   return par[0]*syst_indx_bkg[int(par[2])] +
          par[1]*data_indx_bkg[int(par[2])];
}
	
add_slang_function("fcorfile",["norm","cornorm","id"]);

%%%%%%%%%%%%%%%%%%%%%%%%%%

define fcorfile_default(i)
{
   switch(i)
   {
    case 0:
    return(1,0,0,2);
   }
   {
    case 1:
    return(1,0,0,2);
   }
   {
    case 2:
    return(1,1,1,13);
   }
}

set_param_default_hook("fcorfile","fcorfile_default");

%%%%%%%%%%%%%%%%%%%%%%%

public define corfile()
{
   variable indx, cindx, rscl, rb, n;
   switch(_NARGS)
   {
    case 1:
      indx = ();
      indx = int(abs(indx));

      rb=get_data_info(indx).rebin;
      n=get_data_info(indx).notice_list;

      back_fun(indx,NULL);
      if( syst_indx_app[indx] == 0 )
      {
         set_rebin_error_hook(indx,NULL);
      }
      corback_comp[indx] = 0;
      data_indx_app[indx] = 0;
      rebin_data(indx,rb);
      ignore(indx);
      notice_list(indx,n);
      return;
   }
   {
    case 3:
      (indx, cindx, rscl) = ();

      indx = int(abs(indx));
      cindx = int(abs(cindx));

      back_fun(indx,NULL);

      if( back_indx_app[indx] == 0 )
      {
        create_back(indx);
        syst_indx_bsc[indx] = create_backscale(indx);
        back_indx_app[indx] = 1;
      }
      corback_comp[indx] = 1;

      variable data, ers, bgd, scl;

      rebin_data(cindx,0);
      data = get_data_counts(cindx);
      bgd = get_back(cindx);

      if(bgd == NULL)
      {
         bgd = Float_Type[length(data.value)];
      }

      scl = create_backscale(cindx);

      variable meth = strtok(get_fit_statistic,";=");
      meth = meth[length(meth)-1];

      if( meth == "gehrels" )
      {
         ers = (1.+sqrt(data.err^2+0.75))^2 + scl*bgd;
      }
      else
      {
         ers = data.err^2 + scl*bgd;
      }

      data_indx_bkg[indx] = rscl*(data.value-bgd);
      data_indx_err[indx] = rscl*ers;
      data_indx_app[indx] = 1;

      exclude(cindx);

      back_fun(indx,"fcorfile("+string(indx)+")");    
      set_par("fcorfile("+string(indx)+").id",indx);

      return;
   }
   {
      variable str = [" corfile(data_indx[,corr_indx, scale]);"," ",
       "    Sets both the background associated with data set `data_indx', as well as",
       "    the background subtracted data associated with data set `corr_indx',",
       "    as background correction files for data set `data_indx'.  The latter",
       "    data set is further scaled by a user supplied constant given by `scale'",
       "    (usually the ratio of data_indx over corr_indx exposure times).  Both of",
       "    these components can be scaled further via fit constants using the auto-",
       "    defined function `fcorfile(data_indx)'.  The background errors for both",
       "    of these components will be added in quadrature (either with Poisson or",
       "    Gehrels statistics - whichever was in place the last time the data was",
       "    rebinned), but without accounting for the background fit constants.",
       "    The data set associated with `corr_indx' will be set to be excluded from",
       "    the fits.  No error checking is done to make sure that that data set is",
       "    compatible with the data set from `data_indx'."," ",
       "    To clear the background correction do:", "",
       "       isis> corfile(data_indx)"];
      msg(str);
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Data Functions, Last Updated: Feb. 6, 2008

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Public Functions in This File.  Usage message (almost always) when 
% function is called without arguments.

% list_all      : List all the data, arfs, and rmfs
% load_all      : Loads data & backgrounds & arfs.
% read_radio    : Read radio data from an ascii file, load to struct.
% load_radio    : Take the structure above, and load it into data
% read_optical  : Read (optical) data from file where the units are A_lo,
%                 A_hi, mJy, Delta mJy
% clear_all     : Wipe out all data, arfs, and rmfs

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define list_all() 
{ 
   list_data;
   list_rmf;
   list_arf; 
}

public define load_all(a)
{
   variable b = load_data(a);
   a=a+"[SPECTRUM]";
   variable d = fits_read_key(a,"backfile");
   if(d != NULL and d !="none")
   {
      () = define_back(b,d); 
   }
   else
   {
	() = printf(" %s: No associated background file \n",a);
   }
   return b;
}

%
% Stolen from John Houck & Glenn Allen
%

static variable c_cgs = 2.99792458e10;
static variable h_cgs = 6.62606876e-27;
static variable mjy2cgs = 1.0e-26;

public define read_radio( )
{
   variable in_fil, nbin;
   switch(_NARGS)
   { 
    case 2:
      (in_fil, nbin) = ();
   }
   {
      variable str = [" radio_data = read_radio(radio_file,nbins);",
       " ",
       "   Takes data in the ASCII file `radio_file', comprised of three",
       "   columns: Hz, mJy, \Delta mJy, and places them in a data structure", 
       "   suitable for loading via `load_radio'.  nbins gives the",
       "   number of logarithmically spaced wavelength bins to cover the",
       "   radio data."];
      msg(str);
      return;
   }

   variable in_filp = fopen ( in_fil, "r" ); 

   if ( in_filp == NULL )
   {
      verror ( "\n  ***** Error: Unable to open file %s.\n", in_fil );
   }

   variable frequency, flux_density, flux_density_err;

   ( frequency, flux_density, flux_density_err ) = readcol ( in_fil, 1, 2, 3 );

   () = fclose ( in_filp ); 

   variable n_data_pnts = length ( frequency );

   % Turn Hz into Angstrom, & make sure data is sorted

   variable wavelength = c_cgs / frequency * 1.e8; 

   variable sort_indices = array_sort( wavelength );
   wavelength = wavelength[ sort_indices ]; 

   % Convert to ph cm^-2 s^-1 Hz^-1

   variable flux_differential = flux_density*mjy2cgs/h_cgs/frequency;
   variable flux_differential_err = flux_density_err*mjy2cgs/h_cgs/frequency;

   flux_differential = flux_differential[ sort_indices ];
   flux_differential_err = flux_differential_err[ sort_indices ];

   % Create wavelength grid

   variable wavelength_lo, wavelength_hi;
   variable n_wavelength_bins = nbin;
   variable scal_factor = 1. + (10./double(n_wavelength_bins));

   ( wavelength_lo, wavelength_hi ) = linear_grid( 
                                         log10(min(wavelength)/scal_factor),
                                         log10(max(wavelength)*scal_factor),
                                         n_wavelength_bins 
                                                 );
                                               
   wavelength_lo  = 10.0^wavelength_lo;
   wavelength_hi  = 10.0^wavelength_hi;

   variable flux_integrated     = Double_Type [ n_wavelength_bins ];
   variable flux_integrated_err = Double_Type [ n_wavelength_bins ];

   variable i1min = 0; 
   variable i1max = n_data_pnts;
   variable i1, i2;

   for ( i1 = i1min; i1 < i1max; i1++ )
   {
      i2 = where ( (wavelength_lo <= wavelength[i1]) and
                     (wavelength[i1] < wavelength_hi) );
        
      if( length(i2) > 0 )
      {
        
         %  ph cm^-2 s^-1 (in bin)	     

	flux_integrated[i2] =  flux_differential[i1] * c_cgs * 1.e8 *
	                       (wavelength_hi[i2] - wavelength_lo[i2]) 
                               / wavelength_lo[i2] / wavelength_hi[i2];

	flux_integrated_err[i2] =  flux_differential_err[i1] * c_cgs * 1.e8 *
	                       (wavelength_hi[i2] - wavelength_lo[i2]) 
                               / wavelength_lo[i2] / wavelength_hi[i2];
      }
   }

   % Remove unnecessary bins

   variable flag  = 1;
   i1    = -1;
   variable index = 0;

   while (flag)
   {
      i2 = where ( flux_integrated[[i1+1:]] > 0.0 ) + i1 + 1;
      if ( length(i2) > 0 )
      {
         if ( i2[0] > i1+1 )
	 {
            flux_integrated[index]     = 0.0;
            flux_integrated_err[index] = 0.0;
            wavelength_lo[index]       = wavelength_lo[i1+1];
            wavelength_hi[index]       = wavelength_hi[i2[0]-1];

            i1 = i2[0] - 1;
            index++;
         }
         else
         {
            flux_integrated[index]     = flux_integrated[i2[0]];
            flux_integrated_err[index] = flux_integrated_err[i2[0]];
            wavelength_lo[index]       = wavelength_lo[i2[0]];
            wavelength_hi[index]       = wavelength_hi[i2[0]];

            i1 = i2[0];
            index++;
         }
      }
      else flag = 0;
   }

   i1max = index - 1;
   i1min = where( flux_integrated > 0.0 )[0];
   variable n_out_pnts = i1max - i1min + 1;

   variable radio_data = struct{ bin_lo, bin_hi, value, err };

   radio_data.bin_lo = Double_Type [ n_out_pnts ];
   radio_data.bin_lo = wavelength_lo[[i1min:i1max]];

   radio_data.bin_hi = Double_Type [ n_out_pnts ];
   radio_data.bin_hi = wavelength_hi[[i1min:i1max]];

   radio_data.value  = Double_Type [ n_out_pnts ];
   radio_data.value  = flux_integrated[[i1min:i1max]];

   radio_data.err    = Double_Type [ length(wavelength_lo) ];
   radio_data.err    = flux_integrated_err[[i1min:i1max]];

   radio_data.value = radio_data.value;
   radio_data.err = radio_data.err;

   return radio_data;
}

public define load_radio()
{
   switch(_NARGS)
   {
    case 1:
      variable radio_struct;
      radio_struct = ();
      variable radio_id = define_counts(radio_struct);
      set_data_exposure(radio_id,1.);
      ignore_list(radio_id, where ( radio_struct.value == 0.));
      return radio_id;
   }
   {
      variable str = ["radio_id = load_radio(radio_struct);"," ",
       "   Takes the structure output from `read_radio();' and loads",
       "   it up for fitting.  Also ignores all the zero count",
       "   channels.  (Note: Those could come back if any further filters",
       "   are applied.)"];
      msg(str);
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define read_optical( )
{
   variable in_fil, nbin;
   switch(_NARGS)
   { 
    case 1:
      (in_fil) = ();
   }
   {
      variable str = [" id = read_optical(optical_file);",
       " ",
       "   Takes data in the ASCII file `optical_file', comprised of four",
       "   columns: A_lo, A_hi, mJy, \Delta mJy, and defines a data set with",
       "   identification id.  This will fail horribly for overlapping data bins."];
      msg(str);
      return;
   }

   variable in_filp = fopen ( in_fil, "r" ); 

   if ( in_filp == NULL )
   {
      verror ( "\n  ***** Error: Unable to open file %s.\n", in_fil );
   }

   variable alo, ahi, flux_density, flux_density_err;

   ( alo, ahi, flux_density, flux_density_err ) = readcol ( in_fil, 1,2,3,4 );

   () = fclose ( in_filp ); 

   variable isort = array_sort(alo);

   variable f_conv = 1.50918896;

   alo = alo[isort];
   ahi = ahi[isort];
   flux_density = flux_density[isort];
   flux_density_err = flux_density_err[isort];

   variable i=0, lalo = length(alo), lotemp, hitemp;
   loop(lalo-1)
   {
      if(alo[i+1] != ahi[i]);
      {
        lotemp = [alo[[0:i]],ahi[i],alo[[i+1:]]];
        hitemp = [ahi[[0:i]],alo[i+1],ahi[[i+1:]]];
	alo = lotemp;
        ahi = hitemp;
        flux_density = [flux_density[[0:i]],0.,flux_density[[i+1:]]];
        flux_density_err = [flux_density_err[[0:i]],0.,flux_density_err[[i+1:]]];
        i++;
      }
      i++;
   }

   variable flux = flux_density * f_conv * log( ahi/alo );
   variable eflux = flux_density_err * f_conv * log( ahi/alo);
   variable id = define_counts(alo,ahi,flux,eflux);
   set_data_exposure(id,1);
   ignore_list(id, where ( flux == 0.));
   return id;
}

public define clear_all()
{
   delete_data(all_data);
   delete_arf(all_arfs);
   delete_rmf(all_rmfs);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Some useful fit functions, last Updated: Nov. 19, 2008

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Public Functions in This File

% simple_gpile_fit : Gratings Pileup correction function
%                    (New, better version than in "ABC Guide to Pileup")

% The following work presuming data x-axis follows keV (i.e., Hz -> keV
% for PSD):

% qpo_fit        : Lorentzian, normalized to QPO rms (counts/bin)
% zfc_fit        : Zero-frequency centered Lorentzian (counts/bin)
% plaw_fit       : Counts/bin power-law
% bkn_plaw_fit   : Counts/bin broken power-law
% dbkn_plaw_fit  : Counts/bin doubly broken power-law
% rms_gauss_fit  : Counts/bin Guassian, normalized to RMS.
% sinwave_fit    : Counts/bin sine wave

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%
%% QPO (Lorentzian) %% 
%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%

% Define a s-lang qpo that works better. Properly averages over bin,
% and thus yields a smoother fit.

define qpo_fit(lo,hi,par)
{
   variable l,r,q,f,al,ah;

   % Go from Angstrom to keV (which we pretend is Fourier Hz),
   % Hence it's F[(al-f)/f] - F[(ah-f)/f] below...

   al = _A(lo);
   ah = _A(hi);

   q = par[1];
   f = par[2];

   % In this formulation, par[0] is rms amplitude

   r = par[0]/(0.5 - atan(-2.*q)/PI);

   l = r^2/(al-ah)/PI * 
       ( atan(2.*q*(al-f)/f) - atan(2.*q*(ah-f)/f) );

   l = reverse(l);

   return l;
}

add_slang_function("qpo",["norm [rms]","Q [f/FWHM]","f [Hz]"]);

define qpo_defaults(i)
{
   switch(i)
   {case 0:
      return (0.1,0,0.,1.e8);
   }
   {case 1:
      return (1,0,0.01,1.e3);
   }
   {case 2:
      return (1,0,0.,1.e4);
   }
}

set_param_default_hook("qpo","qpo_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Zero Freq. Centered Lorentzian  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define zfc_fit(lo,hi,par)
{
   variable a,f,al,ah,l;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),
   % Hence it's atan(al/f) - atan(ah/f) below...

   al = _A(lo);
   ah = _A(hi);
   f = par[1];
   
   % In this formulation, par[0] is RMS

   a = par[0]^2 * PI/2./f;

   l = a*f/(al-ah) *  ( atan(al/f) - atan(ah/f) );

   l = reverse(l);

   return l;
}

add_slang_function("zfc",["norm [rms]","f [Hz]"]);

define zfc_defaults(i)
{
   switch(i)
   {case 0:
      return (0.1,0,0.,1.e8);
   }
   {case 1:
      return (1,0,1.e-6,1.e3);
   }
}

set_param_default_hook("zfc","zfc_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Counts/Bin Powerlaw  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%

define plaw_fit(lo,hi,par)
{
   variable a,s,al,ah,l;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),

   al = _A(lo);
   ah = _A(hi);

   a = par[0];
   s = par[1];
   
   if(s != -1)
   {
      l = a/(s+1) * (al^(s+1) - ah^(s+1));
   }
   else
   {
      l = a * ( log(al) - log(ah) );
   }

   l = reverse(l/(al - ah));

   return l;
}

add_slang_function("plaw",["norm","slope"]);

define plaw_defaults(i)
{
   switch(i)
   {case 0:
      return (0.1,0,0.,1.e8);
   }
   {case 1:
      return (-2,0,-5,5);
   }
}

set_param_default_hook("plaw","plaw_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Broken Counts/Bin Powerlaw  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define bkn_plaw_fit(lo,hi,par)
{
   variable a,sa,b,sb,al,ah,aa,l;
   variable iwl,iwh,liwl,liwh;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),

   al = _A(lo);
   ah = _A(hi);
   aa = (al+ah)/2.;
   l = @al;

   a = par[0];
   sa = par[1];
   b = par[2];
   sb = par[3];

   iwl = where(aa<= b);
   iwh = where(aa > b);

   liwl = length(iwl);
   liwh = length(iwh);
   
   if(liwl != 0)
   {
      l[iwl] = a * aa[iwl]^sa;
   }

   if(liwh != 0)
   {
      l[iwh] = a * b^sa * ( aa[iwh]/b )^sb;
   }

   l = reverse(l);

   return l;
}

add_slang_function("bkn_plaw",["norm","slope_1","f_break [Hz]","slope_2"]);

define bkn_plaw_defaults(i)
{
   switch(i)
   {case 0:
      return (0.1,0,0.,1.e8);
   }
   {case 1:
      return (0,0,-5,5);
   }
   {case 2:
      return (1,0,0.,1.e4);
   }
   {case 3:
      return (-2,0,-5,5);
   }
}

set_param_default_hook("bkn_plaw","bkn_plaw_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Double Broken Counts/Bin Powerlaw  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define dbkn_plaw_fit(lo,hi,par)
{
   variable a,sa,ba,sb,bb,sc,al,ah,aa,l;
   variable iwl,iwm,iwh,liwl,liwm,liwh;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),

   al = _A(lo);
   ah = _A(hi);
   aa = (al+ah)/2.;
   l = @al;

   a = par[0];
   sa = par[1];
   ba = par[2];
   sb = par[3];
   bb = par[4];
   sc = par[5];

   iwl = where(aa <= ba);
   iwm = where(aa > ba and aa <= bb);
   iwh = where(aa > bb);

   liwl = length(iwl);
   liwm = length(iwm);
   liwh = length(iwh);
   
   if(liwl != 0)
   {
      l[iwl] = a * aa[iwl]^sa;
   }

   if(liwm != 0)
   {
      l[iwm] = a * ba^sa * ( aa[iwm]/ba )^sb;
   }

   if(liwh != 0)
   {
      l[iwh] = a * ba^sa * (bb/ba)^sb * (aa[iwh]/bb)^sc;
   }

   l = reverse(l);

   return l;
}

add_slang_function("dbkn_plaw",["norm","slope_1","f_brk_1 [Hz]",
                   "slope_2","f_brk_2 [Hz]","slope_3"]);

define dbkn_plaw_defaults(i)
{
   switch(i)
   {case 0:
      return (0.1,0,0.,1.e8);
   }
   {case 1:
      return (0,0,-5,5);
   }
   {case 2:
      return (1,0,0.,1.e4);
   }
   {case 3:
      return (-1,0,-5,5);
   }
   {case 4:
      return (10,0,0.,1.e4);
   }
   {case 5:
      return (-2,0,-5,5);
   }
}

set_param_default_hook("dbkn_plaw","dbkn_plaw_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Counts/Bin Gaussian Normalized to RMS  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#ifexists erf

define rms_gauss_fit(lo,hi,par)
{
   variable r,f,sig,al,ah,l;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),

   al = _A(lo);
   ah = _A(hi);

   r = par[0];
   f = par[1];
   sig = sqrt(2.)*par[2];

   l = r^2/2 * ( erf( (al-f)/sig ) - erf( (ah-f)/sig ) );
   l = reverse(l/(al-ah));

   return l;
}

add_slang_function("rms_gauss",["norm [rms]","f_0 [Hz]","sigma [Hz]"]);

define rms_gauss_defaults(i)
{
   switch(i)
   {case 0:
      return (0.1,0,0.,1.e8);
   }
   {case 1:
      return (1,0,0,1.e4);
   }
   {case 2:
      return (0.2,0,1.e-4,1.e3);
   }
}

set_param_default_hook("rms_gauss","rms_gauss_defaults");

#endif

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Counts/bin Sine Wave %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%

define sinwave_fit(lo,hi,par)
{
   variable a,f,phs,al,ah,l;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),

   al = _A(lo);
   ah = _A(hi);

   a = par[0];
   f = par[1];
   phs = par[2];

   l = a * ( -cos(2*PI*(f*al-phs)) + cos(2*PI*(f*ah-phs)) )/2/PI/f/(al-ah);
   l = reverse(l);

   return l;
}

add_slang_function("sinwave",["norm","freq","phase"]);

define sinwave_defaults(i)
{
   switch(i)
   {case 0:
      return (0.1,0,0.,1.e8);
   }
   {case 1:
      return (1,0,0,1.e4);
   }
   {case 2:
      return (0,0,0.,1.);
   }
}

set_param_default_hook("sinwave","sinwave_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Gratings Pileup, Take II  (Better Version %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define simple_gpile_fit(lo,hi,par,fun)
{  
   % Peak pileup correction goes as:
   %    exp(log(1-pfrac)*[counts/max(counts)])

   variable pfrac = par[0];

   % Pileup scales with model counts from *data set* indx ...
   variable indx = typecast(par[1],Integer_Type);

   if( indx == 0 or pfrac == 0. )
   {
      return fun;   % Quick escape for no changes ...
   }

   % ... but the arf index could be a different number, so get that

   variable arf_indx = get_data_info(indx).arfs;

   % Get arf information

   variable arf = get_arf(arf_indx[0]);

   % In dither regions (or bad pixel areas), counts are down not
   % from lack of area, but lack of exposure.  Pileup fraction
   % therefore should scale with count rate assuming full exposure.
   % Use the arf "fracexpo" column to correct for this effect

   variable fracexpo = get_arf_info(arf_indx[0]).fracexpo;
   if( length(fracexpo) > 1 )
   {
      fracexpo[where(fracexpo ==0)] = 1.;
   }
   else if( fracexpo==0 )
   {
      fracexpo = 1;
   }

   % Rebin arf to input grid, correct for fractional exposure, and 
   % multiply by "fun" to get ("corrected") model counts per bin

   variable mod_cts;
   mod_cts = fun * rebin( lo, hi,
                          arf.bin_lo, arf.bin_hi,
                          arf.value/fracexpo*(arf.bin_hi-arf.bin_lo) )
                / (hi-lo);

   % Go from cts/bin/s -> cts/angstrom/s 

   mod_cts = mod_cts/(hi-lo);

   % Use 2nd and 3rd order arfs to include their contribution.
   % Will probably work best if one chooses a user grid that extends
   % from 1/3 of the minimum wavelength to the maximum, and has
   % at least 3 times the resolution of the first order grid.

   variable mod_ord;
   if(par[2] > 0)
   {
      indx = typecast(par[2],Integer_Type);
      arf = get_arf(indx);
      fracexpo=get_arf_info(indx).fracexpo;
      if( length(fracexpo) > 1 )
      {
         fracexpo[where(fracexpo ==0)] = 1.;
      }
      else if( fracexpo==0 )
      {
         fracexpo = 1;
      }
      mod_ord = arf.value/fracexpo*
                rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
      mod_ord = rebin(lo,hi,2*arf.bin_lo,2*arf.bin_hi,mod_ord)/(hi-lo);
      mod_cts = mod_cts+mod_ord;
   }
   if(par[3] > 0)
   {
      indx = typecast(par[3],Integer_Type);
      arf = get_arf(indx);
      fracexpo=get_arf_info(indx).fracexpo;
      if( length(fracexpo) > 1 )
      {
         fracexpo[where(fracexpo ==0)] = 1.;
      }
      else if( fracexpo==0 )
      {
         fracexpo = 1;
      }
      mod_ord = arf.value/fracexpo*
                rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
      mod_ord = rebin(lo,hi,3*arf.bin_lo,3*arf.bin_hi,mod_ord)/(hi-lo);
      mod_cts = mod_cts+mod_ord;
   }

   variable max_mod = max(mod_cts);
   if(max_mod <= 0){ max_mod = 1.;}

   % Scale maximum model counts to pfrac

   mod_cts = log(1-pfrac)*mod_cts/max_mod;

   % Return function multiplied by exponential decrease

   fun = exp(mod_cts) * fun;
  
   return fun;
}

add_slang_function("simple_gpile",
                   ["pile_frac","data_indx","arf2_indx","arf3_indx"]);

define simple_gpile_defaults(i)
{
   switch(i)
   {case 0:
      return (0.1,1,0.,0.5);
   }
   {case 1:
      return (1,1,0,48);
   }
   {case 2:
      return (2,1,0,48);
   }
   {case 3:
      return (3,1,0,48);
   }
}

set_param_default_hook("simple_gpile", "simple_gpile_defaults");

set_function_category("simple_gpile", ISIS_FUN_OPERATOR);

%%%%%%%%%%%%%%%%%%%%%%%%
define simple_gpile2_fit(lo, hi, par, fun)
%%%%%%%%%%%%%%%%%%%%%%%%
{  
   % 2007, August 14 - fracexpo does not have to be an array

   % 2007, May 03 - correct rebinning of the arf

   % 2007, January 22 - no explicit refering to max(mod_cts)
   
   % 2005, October 25 - New and improved functionality, especially
   %                    in the dithered regions of the chips!

   % Peak pileup correction goes as:
   %    exp(log(1-pfrac)*[counts/max(counts)])
   %  = exp(- beta * counts )
   variable beta = par[0];

   % Pileup scales with model counts from *data set* indx
   variable indx = typecast(par[1],Integer_Type);

   if( indx == 0 or beta == 0. )
   { return fun; }  % Quick escape for no changes ... 

   % The arf index could be a different number, so get that
   variable arf_indx = get_data_info(indx).arfs;

   % Get arf information
   variable arf = get_arf(arf_indx[0]);

   % In dither regions (or bad pixel areas), counts are down not
   % from lack of area, but lack of exposure.  Pileup fraction
   % therefore should scale with count rate assuming full exposure.
   % Use the arf "fracexpo" column to correct for this effect
   variable fracexpo = get_arf_info(arf_indx[0]).fracexpo;
   if(length(fracexpo)>1)
   { fracexpo[where(fracexpo==0)] = 1.; }
   else
   { if(fracexpo==0) { fracexpo = 1; } }

   % Rebin arf to input grid, correct for fractional exposure, and 
   % multiply by "fun" to get ("corrected") model counts per bin
   variable mod_cts_int;
   mod_cts_int = fun * rebin(lo, hi,
                             arf.bin_lo, arf.bin_hi,
                             arf.value*(arf.bin_hi-arf.bin_lo)/fracexpo)
                                      / (hi-lo);
   % Go from bin-integrated(ph/cm^2/s) * bin-integrated(cm^2)
   % to cts/s/angstrom
   variable mod_cts = mod_cts_int/(hi-lo);

   % Use 2nd and 3rd order arfs to include their contribution.
   % Will probably work best if one chooses a user grid that extends
   % from 1/3 of the minimum wavelength to the maximum, and has
   % at least 3 times the resolution of the first order grid.
   variable mod_ord;
   if(par[2] > 0)
   {
      indx = typecast(par[2],Integer_Type);
      arf = get_arf(indx);
      fracexpo=get_arf_info(indx).fracexpo;
      if(length(fracexpo)>1)
      { fracexpo[where(fracexpo==0)] = 1.; }
      else
      { if(fracexpo==0) { fracexpo = 1; } }
      mod_ord = arf.value/fracexpo*
                rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
      mod_ord = rebin(lo,hi,2*arf.bin_lo,2*arf.bin_hi,mod_ord)/(hi-lo);
      mod_cts = mod_cts+mod_ord;
   }
   if(par[3] > 0)
   {
      indx = typecast(par[3],Integer_Type);
      arf = get_arf(indx);
      fracexpo=get_arf_info(indx).fracexpo;
      if(length(fracexpo)>1)
      { fracexpo[where(fracexpo==0)] = 1.; }
      else
      { if(fracexpo==0) { fracexpo = 1.; } }
      mod_ord = arf.value/fracexpo*
                rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
      mod_ord = rebin(lo,hi,3*arf.bin_lo,3*arf.bin_hi,mod_ord)/(hi-lo);
      mod_cts = mod_cts+mod_ord;
   }

   % Return function multiplied by exponential decrease
   return exp(-beta*mod_cts) * fun;
}


add_slang_function("simple_gpile2",
                   ["beta [s*A/cts]","data_indx","arf2_indx","arf3_indx"]);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define simple_gpile2_defaults(i)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
   switch(i)
   {case 0:
      return (0.05,1,0,10);
   }
   {case 1:
      return (0,1,0,48);
   }
   {case 2:
      return (0,1,0,48);
   }
   {case 3:
      return (0,1,0,48);
   }
}

set_param_default_hook("simple_gpile2", "simple_gpile2_defaults");

set_function_category("simple_gpile2", ISIS_FUN_OPERATOR);



%%%%%%%%%%%%%%%%%%%%%%%
define use_simple_gpile()
%%%%%%%%%%%%%%%%%%%%%%%
{
  % process arguments
  variable data1, arf2=NULL, arf3=NULL;
  switch(_NARGS)
  { case 1:  data1 = (); }
  { case 2: (data1, arf2) = (); }
  { case 3: (data1, arf2, arf3) = (); }
  { message("usage: use_simple_gpile(data1[, arf2[, arf3]]);"); return; } 

  if(arf2!=NULL) 
  { arf2 = [arf2];
    if(length(arf2)!=length(data1)) { message("error (use_simple_gpile): arrays data1 and arf2 have different length"); return; }
  }
  if(arf3!=NULL)
  { arf3 = [arf3];
    if(length(arf3)!=length(data1)) { message("error (use_simple_gpile): arrays data1 and arf3 have different length"); return; }
  }
 
  variable simple_gpile_fun = "simple_gpile2";
  variable fitFun = get_fit_fun();
  if(substr(fitFun, 1, strlen(simple_gpile_fun)) != simple_gpile_fun)
  { fit_fun(simple_gpile_fun + "(Isis_Active_Dataset, " + fitFun + ")"); }

  % assign model parameters
  variable i;
  foreach i (all_data)
  { 
    variable simple_gpile_instance = simple_gpile_fun + "(" + string(i) + ")";
    set_par(simple_gpile_instance + ".beta", 0, 1);
    set_par(simple_gpile_instance + ".data_indx", 0, 1);
    set_par(simple_gpile_instance + ".arf2_indx", 0, 1);
    set_par(simple_gpile_instance + ".arf3_indx", 0, 1);
  }
  for(i=0; i<length([data1]); i++)
  {
    variable id = [data1][i];
    simple_gpile_instance = simple_gpile_fun + "(" + string(id) + ")";
    set_par(simple_gpile_instance + ".beta", 0.05, 0);
    set_par(simple_gpile_instance + ".data_indx", id, 1);
    if(arf2!=NULL) { set_par(simple_gpile_instance + ".arf2_indx", arf2[i], 1); }
    if(arf3!=NULL) { set_par(simple_gpile_instance + ".arf3_indx", arf3[i], 1); }
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Utility functions for calculating flux. Last Updated: Nov. 11, 2006

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Public Functions in This File.  Usage message (almost always) when 
% function is called without arguments.

% kev_flux       : Model flux density over a given energy span
% kev_flux_pc    : Integrated model flux, given assumed distance in pc
% lambda_flux    : Model flux density over a given wavelength span
% lambda_flux_pc : Integrated model flux, given assumed distance in pc
% eqw            : Calculate a line equivalent width in mA and eV

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Returns model flux within the *keV range* [ylo,yhi], 
% in Units of ergs/cm^2/sec

public define kev_flux ( )
{
   variable id, e_lo, e_hi, l_lo, l_hi, m, e_avg, pflux, eflux;
   switch(_NARGS)
   {
    case 3:
      (id,e_lo,e_hi) = ();
      l_hi = _A(e_lo);        % Convert keV to A, or visa versa
      l_lo = _A(e_hi);        % Convert keV to A, or visa versa
      m = get_model_flux(id);
      pflux = rebin (l_lo, l_hi, m.bin_lo, m.bin_hi, m.value)[0];   
      e_avg = _A(1.)*(1./m.bin_hi+1./m.bin_lo)/2.*1.602e-9;
      eflux = rebin (l_lo, l_hi, m.bin_lo, m.bin_hi, m.value*e_avg)[0];   
      return pflux, eflux;
   }
   {
      variable str = [" kev_flux(indx,minkev,maxkev);"," ",
      "   Returns model flux - photons/cm^2/s and keV/cm^2/s - ",
      "   for data set indx over the keV energy range spanned by",
      "    minkev/maxkev. Flux is in units of ergs/cm^2/sec"];
      msg(str);
      return;
   }   
}

% Returns data flux within the *keV range* [ylo,yhi], 
% in Units of ergs/cm^2/sec

public define kev_data_flux ( )
{
   variable id, ylo, yhi;
   switch(_NARGS)
   {
    case 3:
      (id,ylo,yhi) = ();
      flux_corr(id);
      variable lhi = _A(ylo);
      variable llo = _A(yhi);
      variable m = get_data_flux(id);
      m.value = m.value*_A(1.)*(1./m.bin_hi+1./m.bin_lo)/2.*1.602e-9;
      m.err = m.err*_A(1.)*(1./m.bin_hi+1./m.bin_lo)/2.*1.602e-9;
      variable flux = rebin (llo, lhi, m.bin_lo, m.bin_hi, m.value)[0];
      variable fluxerr = sqrt(rebin (llo, lhi, m.bin_lo, m.bin_hi, m.err^2)[0]);
      variable n = length( where( m.bin_lo >= llo and m.bin_hi <= lhi) );
      if( n<=1 ) n=2;
      return flux, fluxerr, fluxerr/sqrt(n-1.);
   }
   {
      variable str = [" kev_data_flux(indx,minkev,maxkev);"," ",
      "   Returns data flux, error, and error on mean for data set indx",
      "   over the keV energy range spanned by minkev/maxkev."," ",
      "   Flux is in units of ergs/cm^2/sec"];
      msg(str);
      return;
   }   
}

% Returns model flux within the *keV range* [ylo,yhi], 
% in Units of ergs/sec, for a given distance in pc

define kev_flux_pc ( )
{
   variable id, ylo, yhi, pc;
   switch(_NARGS)
   {
    case 4:
      (id,ylo,yhi,pc) = ();
      variable lhi = _A(ylo);
      variable llo = _A(yhi);
      variable m = get_model_flux(id);
      m.value = m.value*_A(1.)*(1./m.bin_hi+1./m.bin_lo)/2.*1.602e-9;
      return rebin (llo, lhi, m.bin_lo, m.bin_hi, m.value)[0]
             * (pc*3.085e18)^2 * 4. * PI;   
   }
   {
      variable str = [" kev_flux_pc(indx,minkev,maxkev,pc);"," ",
      "   Returns model luminosity for data set indx over the keV energy",
      "   range spanned by minkev/maxkev, presuming a distance of pc parsec",
      " ","   Luminosity is isotropic in units of ergs/sec"];
      msg(str);
      return;
   }   
}

% Returns model flux within the *wavelength range* [ylo,yhi], 
% in Units of ergs/sec for a distance in pc

public define lambda_flux_pc ( )
{
   variable id, ylo, yhi, pc;
   switch(_NARGS)
   {
    case 4:
      return;
      variable m = get_model_flux(id);
      m.value = m.value*_A(1.)*(1./m.bin_hi+1./m.bin_lo)/2.*1.602e-9;
      return rebin (ylo, yhi, m.bin_lo, m.bin_hi, m.value)[0]
                  * (pc*3.085e18)^2 * 4. * PI;
   }
   {
      variable str = [" lambda_flux_pc(indx,minlam,maxlam,pc);"," ",
      "   Returns model luminosity for data set indx over the Angstrom",
      "   range spanned by minlam/maxlam, presuming a distance of pc parsec",
      " ","   Luminosity is isotropic in units of ergs/sec"];
      msg(str);
      return;
   }   
}

% Returns model flux within the *wavelength range* [ylo,yhi], 
% in Units of ergs/cm^2/sec

public define lambda_flux ( )
{
   variable id, ylo, yhi;
   switch(_NARGS)
   {
    case 3:
      return;
      variable m = get_model_flux(id);
      m.value = m.value*_A(1.)*(1./m.bin_hi+1./m.bin_lo)/2.*1.602e-9;
      return rebin (ylo, yhi, m.bin_lo, m.bin_hi, m.value)[0];
   }
   {
      variable str = [" lambda_flux(indx,minlam,maxlam);"," ",
      "   Returns model flux for data set indx over the Angstrom",
      "   range spanned by minlam/maxlam.",
      " ","   Flux is in units of ergs/cm^2/sec"];
      msg(str);
      return;
   }   
}

%%%%%%%%%%%%%%%%%%%

public define eqw()
{
   variable indx, par, pmin=0, lo=NULL, hi=NULL, eva, eve;
   variable fv_hold = Fit_Verbose, noeval=0;

   switch(_NARGS)
   { case 2:
      (indx, par) = ();
   }
   { case 3:
      (indx, par, pmin) = ();
   }
   { case 4:
      (indx, par, pmin, lo) = ();
   }
   { case 5:
      (indx, par, pmin, lo, hi) = ();
   }
   { case 6:
      (indx, par, pmin, lo, hi, noeval) = ();
   }
   {
      variable str = [
      "(ew_ma, ew_ev) = eqw(indx, par, [pmin,] [lo, hi], [noeval]); "," ",
      "  Calculate the equivalent width for a model component, in milli-angstrom and eV.",
      "  ",
      "  Inputs:",
      "   indx    : Data set index to which the model is applied",
      "   par     : Number for, *or* string with the name of, the *normalization*",
      "             parameter for the model component for which the EW will be calculated",
      " ",
      "  Optional Inputs:",
      "   pmin    : For parameter par, what value it should be set to when calculating",
      "             the continuum flux without the line (default = 0, but this allows ",
      "             other parameter toggles to be used).",
      "   lo, hi  : *Wavelength* ranges over which to restrict the",
      "             evaluation of the equivalent width. (Otherwise, uses",
      "             the full wavelength range of the arf; see get_model_flux();)"," ",
      "   noeval  : If = 1, don't do the final eval_counts (parameters will be reset",
      "             to initial values, but not evaluated - i.e., plots won't show",
      "             the fit with the line, and may otherwise be screwy)"," ",
      "  Outputs:",
      "   ew_ma, ew_ev : Equivalent widths in milli-angstrom and eV"];
      msg(str);  
      return;
   }

   if( pmin == NULL )
   {
      pmin = 0;
   }

   Fit_Verbose = -1;
   () = eval_counts;

   variable a = get_model_flux(indx);

   variable phold = get_par_info(par);

   set_par(par,pmin,0,0,phold.max);
   () = eval_counts;
   variable b = get_model_flux(indx);

   set_par(par,phold.value,phold.freeze,phold.min,phold.max);
   if(noeval !=0)
   {
      () = eval_counts;
   }

   variable iw, iwa, iwe;
   if( lo != NULL and hi != NULL )
   {
      iw = where(a.bin_lo >= lo and a.bin_hi <= hi);
   }
   else
   {
      iw = [0:length(a.bin_lo)-1];
   }
   
   variable diff = a.value - b.value;
   variable bdena = b.value/(a.bin_hi-a.bin_lo);
   variable bdene = reverse(reverse(b.value)/(_A(a.bin_lo)-_A(a.bin_hi)));
   variable fdena = diff/(a.bin_hi-a.bin_lo);
   variable fdene = reverse(reverse(diff)/(_A(a.bin_lo) - _A(a.bin_hi)));
   variable flux = sum( diff[iw] );

   iwa = min(where( abs(fdena[iw]) == max(abs(fdena[iw])) ));
   iwe = min(where( abs(fdene[iw]) == max(abs(fdene[iw])) ));
   
   eva = flux/bdena[iw[iwa]]*1000.;
   eve = flux/bdene[iw[iwa]]*1000.;

   Fit_Verbose = fv_hold;

   return eva, eve;
}

