
% Script saved as psd_demo.sl, which can be run either as:
% isis> .load new_isis_script.sl
%  ... or ...
% isis> () = evalfile("psd_demo.sl");

% Utility functions for PSD fitting

()=evalfile("sitar.sl");
()=evalfile("isis_fancy_plots.sl");
()=evalfile("isis_utility_functions.sl");
()=evalfile("isis_utility_functions_prerelease_1.5.sl");

() = printf("\n");
() = printf(" Load script files that will help us perform and plot these analyses: \n");
() = printf("\n");
() = printf("   ()=evalfile(\"sitar.sl\");\n");
() = printf("   ()=evalfile(\"isis_fancy_plots.sl\");\n");
() = printf("   ()=evalfile(\"isis_utility_functions.sl\");\n");
() = printf("   ()=evalfile(\"isis_utility_functions_prerelease_1.5.sl\");\n");

% Declare variables to be used in the script

variable ta,ca,pa,na,ctsa,f,cts,aflo,afhi,apsd,nf;

% Read in the lightcurve.  Convert the counts to integers to save on 
% space and speed up the FFTs a bit.  (An RXTE lightcurve, mostly 1's
% and 0's at this resolution.)

() = printf("\n");
() = printf(" Read and plot FITS file lightcurve:\n");
() = printf("\n");
() = printf("   (ta,ca) = fits_read_col(\"events_18_39_a.lc\",\"time\",\"counts\");\n");

(ta,ca) = fits_read_col("events_18_39_a.lc","time","counts");
ca = typecast(ca,Integer_Type);

xlabel("Time  (sec)");
ylabel("Counts");
plot_bin_integral;
xlin; ylin;

() = printf("   plot(ta-ta[0],ca);\n");
() = printf("\n");

xrange; yrange;
plot(ta-ta[0],ca);

% Calculate the average PSD from 16 sec segments (== 65536 * 2^{-12}).
% The 4th argument are times of arrival, so data gaps can be skipped.
% The 5th argument has been skipped, so we default to the "Leahy
% normalization", where Poisson noise = 2

() = printf(" Create average PSD from 16 sec segments:\n");
() = printf("\n");
() = printf("   (f,pa,na,ctsa) = sitar_avg_psd(ca,65536,1./2^12,ta);\n");
() = printf("\n");
  
(f,pa,na,ctsa) = sitar_avg_psd(ca,65536,1./2^12,ta);  

% OK, let's bin that up a bit, by logarithmically averaging over bins
% that are 1% wide (Delta f/f).

(aflo,afhi,apsd,nf) = sitar_lbin_psd(f,pa,0.01); 

() = printf(" Logarithmically bin the PSD with 0.01 bins:\n");
() = printf("\n");
() = printf("   (aflo,afhi,apsd,nf) = sitar_lbin_psd(f,pa,0.01);\n");
() = printf("\n");
() = printf(" Plot, and measure the QPO frequency with your cursor:\n");
() = printf("\n");
() = printf("   xlabel(\"Frequency (Hz)\"); ylabel(\"PSD\");\n");
() = printf("   hplot(aflo,afhi,apsd);\n");
() = printf("   cursor;\n");
() = printf("\n");

xlabel("Frequency (Hz)\n"); 
ylabel("PSD\n");
hplot(aflo,afhi,apsd);
cursor;

% Define the logarithmically averaged PSD to be the data set. PSD error,
% for a stochastic process, is 100%/sqrt(number of averages).  Here, that
% number is number of independent segments averaged over X number of 
% frequencies in each logarithmic bin.  Those numbers are returned by the
% above routines.

() = printf(" Assign the PSD as a fittable data set:\n");
() = printf("\n");
() = printf("   id = sitar_define_psd(aflo,afhi,apsd,apsd/sqrt(na*nf));\n");
() = printf(" \n");

variable id = sitar_define_psd(aflo,afhi,apsd,apsd/sqrt(na*nf));

% Only notice frequencies between 500 and 1500 Hz.  (In my scheme, I equate 
% FFT Hz with spectral unit keV, 1:1.  I.e., in all functions, just pretend
% keV = Hz, hence the "xnotice_en" = exclusively notice energy range 500->1500
% in data set id).

() = printf(" Only notice 500 Hz - 1500 Hz\n");
() = printf("\n");
() = printf("   xnotice_en(id,500,1500);\n");
() = printf("\n");

xnotice_en(id,500,1500);

% Define a fit function.  The "qpo" is from my isis start up files, and not an 
% ISIS intrinsic function.

() = printf(" Define the fit function, then fit:\n");
() = printf("\n");
() = printf("   fit_fun(\"constant(1)+qpo(1)\");\n");
() = printf("\n");
() = printf("   set_par(\"constant(1).factor\",1.99);\n");
() = printf("   set_par(\"qpo(1).norm\",1.8);\n");
() = printf("   set_par(\"qpo(1).Q\",20);\n");
() = printf("   set_par(\"qpo(1).f\",750);\n");
() = printf("\n");
() = printf("   () = renorm_counts;\n");
() = printf("\n");

fit_fun("constant(1)+qpo(1)");

% Set some starting parameters (the Poisson noise level is slightly < 2 due to
% deadtime effects).

set_par("constant(1).factor",1.99);
set_par("qpo(1).norm",1.8);
set_par("qpo(1).Q",20);
set_par("qpo(1).f",750);

% Do the fit

() = renorm_counts;

() = printf("\n");
() = printf("   () = fit_counts;\n");
() = printf("\n");

() = fit_counts;
() = printf("\n");
() = printf("   list_par;\n");
() = printf("\n");

list_par;

() = printf("\n");

plot_pause;

% Plot via my plotting routines

() = printf("\n");
() = printf(" Plot the results:\n");
() = printf("\n");
() = printf("   d_width=3; r_width=3; m_width=3; de_width=1; re_width=1;\n");
() = printf("   set_frame_line_width(3);\n");
() = printf("   charsize(1.12);\n");
() = printf("   Plot_Unit(\"psd_leahy\");\n");
() = printf(" \n");
() = printf("   popt.dsym=0;\n");
() = printf("   popt.dcol=4;\n");
() = printf("   popt.decol=5;\n");
() = printf("   popt.rsym=0;\n");
() = printf("   popt.rcol=4;\n");
() = printf("   popt.recol=5;\n");
() = printf("   popt.xrng=[500,1500];\n");
() = printf("   popt.yrng=[1.9801,2.2];\n");
() = printf("   popt.res=1;\n");
() = printf("\n");

() = printf("   pid = open_print(\"psd_plot.ps/vcps\");\n");
() = printf("      resize(21,0.85);\n");
() = printf("      plot_counts(id,popt);\n");
() = printf("   close_print(pid,\"gv\"); \n");
() = printf("\n");

d_width=3; r_width=3; m_width=3; de_width=1; re_width=1;
set_frame_line_width(3);
charsize(1.12);
Plot_Unit("psd_leahy");

popt.dsym=0;
popt.dcol=4;
popt.decol=5;
popt.rsym=0;
popt.rcol=4;
popt.recol=5;
popt.xrng=[500,1500];
popt.yrng=[1.9801,2.2];
popt.res=1;

variable pid = open_print("psd_plot.ps/vcps");  % open/close_print are my wrappers
resize(21,0.85);    

plot_counts(id,popt);

close_print(pid,"gv");  

() = printf(" Find the error bars, and store in a file:\n");
() = printf("\n");
() = printf("   (,) = conf_loop([1:4];save,prefix=\"qpo\");\n");
() = printf("   () = system(\"more qpo.save\");\n");
() = printf("\n");

(,) = conf_loop([1:4];save,prefix="qpo");
() = system("more qpo.save");

() = printf("\n");
() = printf(" Do a: delete_data(all_data);  to start from scratch.\n");
() = printf("\n");



